Imports System.Net.Sockets
Imports System.IO
Imports System.Threading.Tasks

Public Class TcpEndpoint : Implements IDisposable

   Public Event DataReceived As EventHandler
   Public Event Disposed As EventHandler
   Public Event Initialized As EventHandler

   Private _ID As Integer
   Private _CommandCode As New UInt16
   Private _SmallBuf(3) As Byte

   Public ReadOnly TcpClient As TcpClient, WriteStream, ReadStream As StreamContainer
   
   Public Sub New(ByVal TC As TcpClient, Optional id As Integer? = Nothing)
      TcpClient = TC
      Dim strm = TcpClient.GetStream
      ReadStream = New StreamContainer(strm, FileAccess.Read)
      WriteStream = New StreamContainer(strm, FileAccess.Write)
      DoWork(id)
   End Sub
   
   Private Async Sub DoWork(id As Integer?)
      If id.HasValue Then 'indicates, that I'm a servers Endpoint, and initialize dirctly with argument
         _ID = id.Value
      Else 'as client first receive init-message (see it as 'handshake')
         Await task.Run(AddressOf ReceiveInit)
      End If
      While Await task.Run(AddressOf ReceiveContent)
      End While
   End Sub

   ''' <summary> read the command-code from ReadStream </summary>
   Private Function ReceiveContent() As Boolean
      If Not TryReadSmallBuf(2) Then Return False
      _CommandCode = BitConverter.ToUInt16(_SmallBuf, 0)
      RaiseEvent DataReceived(Me, EventArgs.Empty)
      ReadStream.NextContent()
      Return True
   End Function
   ''' <summary> read the ID, generated and sent from Server </summary>
   Private Sub ReceiveInit()
      If Not TryReadSmallBuf(4) Then Return
      _ID = BitConverter.ToInt32(_SmallBuf, 0)
      RaiseEvent Initialized(Me, EventArgs.Empty)
      ReadStream.NextContent()
   End Sub

   Private Function TryReadSmallBuf(count As Integer) As Boolean
      Try
         ReadStream.Read(_SmallBuf, 0, count)
      Catch ex As IOException
         Dispose()
         Return False
      End Try
      Return True
   End Function

   Protected Sub Dispose(ByVal disposing As Boolean)
      If _SmallBuf.Null Then Return
      _SmallBuf = Nothing
      DisposeAll(WriteStream, ReadStream, TcpClient)
   End Sub

   Public Sub Dispose() Implements IDisposable.Dispose
      If _SmallBuf.Null Then Return
      Dispose(True)
      RaiseEvent Disposed(Me, EventArgs.Empty)
      GC.SuppressFinalize(Me)
   End Sub

   Public ReadOnly Property IsDisposed() As Boolean
      Get
         Return _SmallBuf.Null
      End Get
   End Property

   Public ReadOnly Property ID() As Integer
      Get
         Return _ID
      End Get
   End Property

   Public ReadOnly Property CommandCode() As UInt16
      Get
         Return _CommandCode
      End Get
   End Property

   Public Sub WriteCommandCode(cmd As UInt16)
      WriteStream.Write(BitConverter.GetBytes(cmd), 0, 2)
   End Sub

End Class
